from samson.math.polynomial import Polynomial
from samson.math.general import poly_to_int
from samson.prngs.glfsr import GLFSR

class FLFSR(GLFSR):
    """
    Fibonacci linear-feedback shift register.
    """

    def __init__(self, seed: int, polynomial: Polynomial):
        """
        Parameters:
            seed              (int): Initial value.
            polynomial (Polynomial): Either a `Polynomial` or an integer that represents the polynomal.
        """
        self.state = seed

        if type(polynomial) is Polynomial:
            polynomial = poly_to_int(polynomial)

        self.polynomial = polynomial
        self.mask = 1
        self.wrap_around_mask = 2 ** (polynomial.bit_length()) - 1
        self.state &= self.wrap_around_mask


        poly_mask = polynomial
        while poly_mask:
            if poly_mask & self.mask:
                poly_mask ^= self.mask

            if not poly_mask:
                break

            self.mask <<= 1

        self.mask -= 1



    def __reprdir__(self):
        return ['state', 'polynomial', 'mask']



    def clock(self) -> int:
        """
        Generates the next pseudorandom output.

        Returns:
            int: Next pseudorandom output.
        """
        self.state <<= 1

        lsb = sum([int(bit) for bit in bin(self.state & self.polynomial)[2:]]) % 2
        self.state |= lsb
        self.state &= 0xFFFFFFFFFFFFFFFF
        return self.state & 1



    def reverse_clock(self, output: int):
        """
        Clocks the state in reverse given the previous output.

        Parameters:
            output (int): Previous output.
        """
        raise NotImplementedError()



    def generate(self) -> int:
        """
        Calls self.clock(). Here for interface uniformity.

        Returns:
            int: Next pseudorandom output.
        """
        return self.clock()
